/*
 * This file is part of the Score-P software (http://www.score-p.org)
 *
 * Copyright (c) 2009-2012,
 * RWTH Aachen University, Germany
 *
 * Copyright (c) 2009-2012,
 * Gesellschaft fuer numerische Simulation mbH Braunschweig, Germany
 *
 * Copyright (c) 2009-2012, 2014,
 * Technische Universitaet Dresden, Germany
 *
 * Copyright (c) 2009-2012,
 * University of Oregon, Eugene, USA
 *
 * Copyright (c) 2009-2012, 2017,
 * Forschungszentrum Juelich GmbH, Germany
 *
 * Copyright (c) 2009-2012,
 * German Research School for Simulation Sciences GmbH, Juelich/Aachen, Germany
 *
 * Copyright (c) 2009-2012,
 * Technische Universitaet Muenchen, Germany
 *
 * This software may be modified and distributed under the terms of
 * a BSD-style license.  See the COPYING file in the package base
 * directory for details.
 *
 */

#ifndef SCALASCA_ERROR_CODES_H
#define SCALASCA_ERROR_CODES_H


/**
 * @file            SCALASCA_ErrorCodes.h
 * @ingroup         SCALASCA_Exception_module
 *
 * @brief           Error codes and error handling.
 *
 */

#include <errno.h>
#include <stdint.h>
#include <stdarg.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * This is the list of error codes for SCALASCA.
 */
typedef enum
{
    /** Special marker for error messages which indicates a deprecation. */
    SCALASCA_DEPRECATED    = -3,

    /** Special marker when the application will be aborted. */
    SCALASCA_ABORT         = -2,

    /** Special marker for error messages which are only warnings. */
    SCALASCA_WARNING       = -1,

    /** Operation successful */
    SCALASCA_SUCCESS       = 0,

    /** Invalid error code
     *
     * Should only be used internally and not as an actual error code.
     */
    SCALASCA_ERROR_INVALID = 1,

    /* These are the internal implementation of POSIX error codes. */
    /** The list of arguments is too long */
    SCALASCA_ERROR_E2BIG,
    /** Not enough rights */
    SCALASCA_ERROR_EACCES,
    /** Address is not available */
    SCALASCA_ERROR_EADDRNOTAVAIL,
    /** Address family is not supported */
    SCALASCA_ERROR_EAFNOSUPPORT,
    /** Resource temporarily not available */
    SCALASCA_ERROR_EAGAIN,
    /** Connection is already processed */
    SCALASCA_ERROR_EALREADY,
    /** Invalid file pointer */
    SCALASCA_ERROR_EBADF,
    /** Invalid message */
    SCALASCA_ERROR_EBADMSG,
    /** Resource or device is busy */
    SCALASCA_ERROR_EBUSY,
    /** Operation was aborted */
    SCALASCA_ERROR_ECANCELED,
    /** No child process available */
    SCALASCA_ERROR_ECHILD,
    /** Connection was refused */
    SCALASCA_ERROR_ECONNREFUSED,
    /** Connection was reset */
    SCALASCA_ERROR_ECONNRESET,
    /** Resolved deadlock */
    SCALASCA_ERROR_EDEADLK,
    /** Destination address was expected */
    SCALASCA_ERROR_EDESTADDRREQ,
    /** Domain error */
    SCALASCA_ERROR_EDOM,
    /** Reserved */
    SCALASCA_ERROR_EDQUOT,
    /** File does already exist */
    SCALASCA_ERROR_EEXIST,
    /** Invalid address */
    SCALASCA_ERROR_EFAULT,
    /** File is too large */
    SCALASCA_ERROR_EFBIG,
    /** Operation is work in progress */
    SCALASCA_ERROR_EINPROGRESS,
    /** Interruption of an operating system call */
    SCALASCA_ERROR_EINTR,
    /** Invalid argument */
    SCALASCA_ERROR_EINVAL,
    /** Generic I/O error */
    SCALASCA_ERROR_EIO,
    /** Socket is already connected */
    SCALASCA_ERROR_EISCONN,
    /** Target is a directory */
    SCALASCA_ERROR_EISDIR,
    /** Too many layers of symbolic links */
    SCALASCA_ERROR_ELOOP,
    /** Too many opened files */
    SCALASCA_ERROR_EMFILE,
    /** Too many links */
    SCALASCA_ERROR_EMLINK,
    /** Message buffer is too small */
    SCALASCA_ERROR_EMSGSIZE,
    /** Reserved */
    SCALASCA_ERROR_EMULTIHOP,
    /** Filename is too long */
    SCALASCA_ERROR_ENAMETOOLONG,
    /** Network is down */
    SCALASCA_ERROR_ENETDOWN,
    /** Connection was reset from the network */
    SCALASCA_ERROR_ENETRESET,
    /** Network is not reachable */
    SCALASCA_ERROR_ENETUNREACH,
    /** Too many opened files */
    SCALASCA_ERROR_ENFILE,
    /** No buffer space available */
    SCALASCA_ERROR_ENOBUFS,
    /** No more data left in the queue */
    SCALASCA_ERROR_ENODATA,
    /** This device does not support this function */
    SCALASCA_ERROR_ENODEV,
    /** File or directory does not exist */
    SCALASCA_ERROR_ENOENT,
    /** Cannot execute binary */
    SCALASCA_ERROR_ENOEXEC,
    /** Locking failed */
    SCALASCA_ERROR_ENOLCK,
    /** Reserved */
    SCALASCA_ERROR_ENOLINK,
    /** Not enough main memory available */
    SCALASCA_ERROR_ENOMEM,
    /** Message has not the expected type */
    SCALASCA_ERROR_ENOMSG,
    /** This protocol is not available */
    SCALASCA_ERROR_ENOPROTOOPT,
    /** No space left on device */
    SCALASCA_ERROR_ENOSPC,
    /** No stream available */
    SCALASCA_ERROR_ENOSR,
    /** This is not a stream */
    SCALASCA_ERROR_ENOSTR,
    /** Requested function is not implemented */
    SCALASCA_ERROR_ENOSYS,
    /** Socket is not connected */
    SCALASCA_ERROR_ENOTCONN,
    /** This is not a directory */
    SCALASCA_ERROR_ENOTDIR,
    /** This directory is not empty */
    SCALASCA_ERROR_ENOTEMPTY,
    /** No socket */
    SCALASCA_ERROR_ENOTSOCK,
    /** This operation is not supported */
    SCALASCA_ERROR_ENOTSUP,
    /** This IOCTL is not supported by the device */
    SCALASCA_ERROR_ENOTTY,
    /** Device is not yet configured */
    SCALASCA_ERROR_ENXIO,
    /** Operation is not supported by this socket */
    SCALASCA_ERROR_EOPNOTSUPP,
    /** Value is to long for the datatype */
    SCALASCA_ERROR_EOVERFLOW,
    /** Operation is not permitted */
    SCALASCA_ERROR_EPERM,
    /** Broken pipe */
    SCALASCA_ERROR_EPIPE,
    /** Protocol error */
    SCALASCA_ERROR_EPROTO,
    /** Protocol is not supported */
    SCALASCA_ERROR_EPROTONOSUPPORT,
    /** Wrong protocol type for this socket */
    SCALASCA_ERROR_EPROTOTYPE,
    /** Value is out of range */
    SCALASCA_ERROR_ERANGE,
    /** Filesystem is read only */
    SCALASCA_ERROR_EROFS,
    /** This seek is not allowed */
    SCALASCA_ERROR_ESPIPE,
    /** No matching process found */
    SCALASCA_ERROR_ESRCH,
    /** Reserved */
    SCALASCA_ERROR_ESTALE,
    /** Timeout in file stream or IOCTL */
    SCALASCA_ERROR_ETIME,
    /** Connection timed out */
    SCALASCA_ERROR_ETIMEDOUT,
    /** File could not be executed while it is opened */
    SCALASCA_ERROR_ETXTBSY,
    /** Operation would be blocking */
    SCALASCA_ERROR_EWOULDBLOCK,
    /** Invalid link between devices */
    SCALASCA_ERROR_EXDEV,

    /* These are the error codes specific to the Scalasca package */

    /** Memory allocation failed */
    SCALASCA_ERROR_MEM_ALLOC_FAILED,
    /** End of buffer/file reached */
    SCALASCA_ERROR_END_OF_BUFFER,
    /** Invalid file operation */
    SCALASCA_ERROR_FILE_INTERACTION,
    /** Unable to open file */
    SCALASCA_ERROR_FILE_CAN_NOT_OPEN
} SCALASCA_ErrorCode;


/**
 * Returns the name of an error code.
 *
 * @param errorCode : Error Code
 *
 * @returns Returns the name of a known error code, and "INVALID_ERROR" for
 *          invalid or unknown error IDs.
 * @ingroup SCALASCA_Exception_module
 */
const char*
SCALASCA_Error_GetName( SCALASCA_ErrorCode errorCode );


/**
 * Returns the description of an error code.
 *
 * @param errorCode : Error Code
 *
 * @returns Returns the description of a known error code.
 * @ingroup SCALASCA_Exception_module
 */
const char*
SCALASCA_Error_GetDescription( SCALASCA_ErrorCode errorCode );


/**
 * Signature of error handler callback functions. The error handler can be set
 * with @ref SCALASCA_Error_RegisterCallback.
 *
 * @param userData        : Data passed to this function as given by the registration call.
 * @param file            : Name of the source-code file where the error appeared
 * @param line            : Line number in the source code where the error appeared
 * @param function        : Name of the function where the error appeared
 * @param errorCode       : Error code
 * @param msgFormatString : Format string like it is used for the printf family.
 * @param va              : Variable argument list
 *
 * @returns Should return the errorCode
 */
typedef SCALASCA_ErrorCode
( *SCALASCA_ErrorCallback )( void*          userData,
                         const char*    file,
                         uint64_t       line,
                         const char*    function,
                         SCALASCA_ErrorCode errorCode,
                         const char*    msgFormatString,
                         va_list        va );


/**
 * Register a programmers callback function for error handling.
 *
 * @param errorCallbackIn : Function which will be called instead of printing a
 *                          default message to standard error.
 * @param userData :        Data pointer passed to the callback.
 *
 * @returns Function pointer to the former error handling function.
 * @ingroup SCALASCA_Exception_module
 *
 */
SCALASCA_ErrorCallback
SCALASCA_Error_RegisterCallback( SCALASCA_ErrorCallback errorCallbackIn,
                             void*              userData );


#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* SCALASCA_ERROR_CODES_H */
